# =====================================================================
# $Id: swulib.awk 37 2007-05-02 14:41:02Z carlo $
#
# AWK functions for Shell-Ware Utilities.
#
# Copyright (c) 2001,2009 Carlo Strozzi
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License 
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# =====================================================================

# =====================================================================
# string _strip(string s, [int mode])
#
# Strip leading and trailing blanks and TABs from the input string.
# Returns the stripped string. If 'mode' is set to '1', then multiple
# middle blanks are squeezed. If 'mode' is set to '2' then any
# newlines in 'string' are turned to blanks prior to stripping.
# =====================================================================

function _strip(s,m) {
  if (m == 2) gsub(/\n/," ",s)
  sub(/^[ \t]+/,_NULL,s)
  sub(/[ \t]+$/,_NULL,s)
  if (m) gsub(/  +/," ",s)
  return s
}

# =====================================================================
# string _b64decode(string value)
#
# Original code (C) March 10, 2003, Peter van Eerten
# Source: http://www.turtle.dds.nl/b64dec.awk
#
# Modified for SWU by Carlo Strozzi, November 10th, 2006
#
# Take a base64-encoded string and decode it to its original value. 
# Being mawk(1) not 8-bit clean, the original unencoded string should
# contain just text. For the decoding of arbitrary binary objects the
# 'swu-encdec' utility should be used instead of this function.
# =====================================================================

function _b64decode(value,	base1,base2,base3,base4,tmp,\
				byte1,byte2,byte3,byte4,result) {

  result = ""

  while (length(value) > 0) {
	# Specify byte values
	base1 = substr(value, 1, 1)
	base2 = substr(value, 2, 1)
	base3 = substr(value, 3, 1)
	base4 = substr(value, 4, 1)
	# Now find numerical position in _BASE64 string
	byte1 = index(_BASE64, base1) - 1
	if (byte1 < 0) byte1 = 0
	byte2 = index(_BASE64, base2) - 1
	if (byte2 < 0) byte2 = 0
	byte3 = index(_BASE64, base3) - 1
	if (byte3 < 0) byte3 = 0
	byte4 = index(_BASE64, base4) - 1
	if (byte4 < 0) byte4 = 0
	# Reconstruct ASCII string

	# gawk(1) compatibility test.
	if ((tmp=_lshift(_and(byte1, 63), 2) + _rshift(_and(byte2, 48), 4)))
				result = result sprintf("%c", tmp)

	# gawk(1) compatibility test.
	if ((tmp=_lshift(_and(byte2, 15), 4) + _rshift(_and(byte3, 60), 2)))
				result = result sprintf("%c", tmp)

	# gawk(1) compatibility test.
	if ((tmp=_lshift(_and(byte3, 3), 6) + byte4))
				result = result sprintf("%c", tmp)

	# Decrease incoming string by 4
	value = substr(value, 5)
  }
  return result
}

# =====================================================================
# string _b64encode(string value)
#
# Original code (C) March 6, 2003, Peter van Eerten
# Source: http://www.turtle.dds.nl/b64enc.awk
#
# Modified for SWU by Carlo Strozzi, November 10th, 2006
#
# Take a string and return its base64-encoded value. Being mawk(1) not
# 8-bit clean, the original unencoded string should contain just text.
# For the encoding of arbitrary binary objects the 'swu-encdec' utility
# should be used instead of this function.
# =====================================================================

function _b64encode(value,	byte1,byte2,byte3,byte4,\
				base1,base2,base3,base4,result) {
  result = ""

  while (length(value) > 0) {
	# Specify byte values
	if (length(value) == 1){
		byte1 = _asc(substr(value, 1, 1))
		byte2 = 0
		byte3 = 0
	}
	if (length(value) == 2) {
		byte1 = _asc(substr(value, 1, 1))
		byte2 = _asc(substr(value, 2, 1))
		byte3 = 0
	}
	if (length(value) >= 3) {
		byte1 = _asc(substr(value, 1, 1))
		byte2 = _asc(substr(value, 2, 1))
		byte3 = _asc(substr(value, 3, 1))
	}
	# Create BASE64 values
	base1 = _rshift(byte1, 2)
	base2 = _lshift(_and(byte1, 3), 4) + _rshift(_and(byte2, 240), 4)
	base3 = _lshift(_and(byte2, 15), 2) + _rshift(_and(byte3, 192), 6)
	base4 = _and(byte3, 63)
	# Compose BASE64 string
	if (length(value) == 1) {
		result = result substr(_BASE64, base1 + 1, 1)
		result = result substr(_BASE64, base2 + 1, 1)
		result = result "=="
		value = ""
	}
	if (length(value) == 2) {
		result = result substr(_BASE64, base1 + 1, 1)
		result = result substr(_BASE64, base2 + 1, 1)
		result = result substr(_BASE64, base3 + 1, 1)
		result = result "="
		value = ""
	}
	if (length(value) >= 3) {
		result = result substr(_BASE64, base1 + 1, 1)
		result = result substr(_BASE64, base2 + 1, 1)
		result = result substr(_BASE64, base3 + 1, 1)
		result = result substr(_BASE64, base4 + 1, 1)
		value = substr(value, 4)
	}
  }
  return result
}

# =====================================================================
# string _and(string var, string x)
#
# Bitwise AND between 2 variables - var AND x
#
# Original code (C) March 6, 2003, Peter van Eerten
# Source: http://www.turtle.dds.nl/b64enc.awk
#
# Modified for SWU by Carlo Strozzi, November 10th, 2006
# =====================================================================

function _and(var,x,		l_res,l_i) {

  l_res=0;

  for (l_i=0; l_i < 8; l_i++){
        if (var%2 == 1 && x%2 == 1) l_res=l_res/2 + 128;
        else l_res/=2;
        var=int(var/2);
        x=int(x/2);
  }
  return l_res;
}

# =====================================================================
# string _and(string a, string b)
#
# Bitwise AND between 2 variables - a AND b (alternate form).
#
# Original code by A. Costa, Sat Apr 13 2002
# Source: http://www.gis.net/~agcosta/mu_files/awktest.gz
# Bugs: does not understand two's complement negative numbers.
# See http://mulinux.dotsrc.org/archive/8091.html
#
# Modified for SWU by Carlo Strozzi, September 16th, 2008
# =====================================================================
#
#function _and(a,b,    r,n)	# bitwise 'and'
#{
#	r=0	# result
#	n=0	# bit place
#	while (a&&b)
#	{
#		if ( (a*b)%2 )		# both odd?
#			r+=2^n		# set bit 'n'
#		a=int(a/2)
#		b=int(b/2)
#		n++
#	}
#	return r
#}

# =====================================================================
# string _lshift(string var, int x)
#
# Rotate bytevalue left x times
#
# Original code (C) March 6, 2003, Peter van Eerten
# Source: http://www.turtle.dds.nl/b64enc.awk
#
# Modified for SWU by Carlo Strozzi, November 10th, 2006
# =====================================================================

function _lshift(var,x) {

  while(x > 0){
    var*=2;
    x--;
  }
  return var;
}

# =====================================================================
# string _rshift(string var, int x)
#
# Rotate bytevalue right x times
#
# Original code (C) March 6, 2003, Peter van Eerten
# Source: http://www.turtle.dds.nl/b64enc.awk
#
# Modified for SWU by Carlo Strozzi, November 10th, 2006
# =====================================================================

function _rshift(var,x) {

  while(x > 0){
    var=int(var/2);
    x--;
  }
  return var;
}

# =====================================================================
# int _asc(char value)
#
# Return the integer value of a single ASCII character.
#
# Original code (C) March 6, 2003, Peter van Eerten
# Source: http://www.turtle.dds.nl/b64enc.awk
#
# Modified for SWU by Carlo Strozzi, November 10th, 2006
# =====================================================================

function _asc(char,		i) {

  for (i = 32; i < 127; i++){
      if (sprintf("%c", i) == char) return i;
  }
}

# =====================================================================
# string _or(string a, string b)
#
# Bitwise inclusive OR between 2 variables - a OR b
#
# Original code by A. Costa, Sat Apr 13 2002
# Source: http://www.gis.net/~agcosta/mu_files/awktest.gz
# Bugs: does not understand two's complement negative numbers.
# See http://mulinux.dotsrc.org/archive/8091.html
#
# Modified for SWU by Carlo Strozzi, September 16th, 2008
# =====================================================================

function _or(a,b,    r,n) {

  r=0	# result
  n=0	# bit place

  while (a||b) {
  	if ( a%2 || b%2 )	# either odd?
  		r+=2^n		# set bit 'n'
  	a=int(a/2)
  	b=int(b/2)
  	n++
  }
  return r
}

# =====================================================================
# string _xor(string a, string b)
#
# Bitwise XOR between 2 variables - a XOR b
#
# Original code by A. Costa, Sat Apr 13 2002
# Source: http://www.gis.net/~agcosta/mu_files/awktest.gz
# Bugs: does not understand two's complement negative numbers.
# See http://mulinux.dotsrc.org/archive/8091.html
#
# Modified for SWU by Carlo Strozzi, September 16th, 2008
# =====================================================================

function _xor(a,b,    r,n) {

  r=0	# result
  n=0	# bit place

  while (a||b) {
  	if ( (a+b)%2 )		# only one odd?
  		r+=2^n		# set bit 'n'
  	a=int(a/2)
  	b=int(b/2)
  	n++
  }
  return r
}

# =====================================================================
# string _not(string a)
#
# Bitwise NOT function
#
# Original code by A. Costa, Sat Apr 13 2002
# Source: http://www.gis.net/~agcosta/mu_files/awktest.gz
# Bugs: does not understand two's complement negative numbers.
# See http://mulinux.dotsrc.org/archive/8091.html
#
# Modified for SWU by Carlo Strozzi, September 16th, 2008
# =====================================================================

function _not(a,    r,n) {

  r=0	# result
  n=0	# bit place

  while (a) {
  	if ( a%2==0 )		# even?
  		r+=2^n		# set bit 'n'
  	a=int(a/2)
  	n++
  }
  return r
}

# =====================================================================
# int _x2d(string s)
#
# Convert an hexadecimal string to the corresponding integer value.
# The hexadecimal input string is case-insensitive.
#
# Example:  either _x2d("7E") or _x2d("7e") will both return 126 .
#
# Original code by Carlos Duarte <cgd@mail.teleweb.pt>, 22/12/2003 .
# Source: http://cgd.sdf-eu.org/a/libs/awk/hex.awk
#
# Modified for SWU by Carlo Strozzi, September 16th, 2008
# =====================================================================

function _x2d(s,	n,i,j,h) {

  s = tolower(s)
  n = length(s)
  for (i=1; i<=n; i++) {
      j = index("0123456789abcdef", substr(s, i, 1))
      if (j == 0) return -1; 
      h = h*16 + j-1
  }
  return h
}

# =====================================================================
# int _b2d(string b)
#
# Binary to integer conversion.
#
# Convert a string representation of a binary value to an integer.
#
# Example:  _b2d("10101111") will return 175 .
#
# Adapted from shell code at the following URL:
# http://archives.devshed.com/forums/unix-linux-135/ascii-to-decimal-conversion-372898.html
#
# Ported to SWU by Carlo Strozzi, September 16th, 2008
# =====================================================================

function _b2d(b,	d,m,p) {
  do {
     m = substr(b,2)
     p = m "$"
     sub(p,"",b)
     d = (d * 2 + b)
     b = m
  } while (b != "")
  return d
}

# =====================================================================
# int _atoi(string char)
#
# Convert a single character to an integer.
#
# There's no need for a corresponding _itoa(), use printf("%c",int) .
# =====================================================================

function _atoi(char,             i) {

  # I'm currently only interested in the ISO-8859-1 subset.
  for (i=1; i < 256; i++){
      if (sprintf("%c", i) == char) return i;
  }
}

# =====================================================================
# string _u8encode(string s)
#
# Convert a string to UTF-8
#
# Original code from:
# http://www.webtoolkit.info/javascript-url-decode-encode.html
#
# Ported to SWU by Carlo Strozzi, September 16th, 2008
# =====================================================================

function _u8encode(s,	n,i,j,u) {

  i = split(s,a,"")  

  for (j=1; j<=i; j++) {
      n = _atoi(a[j])
      if (n < 128) u = u a[j]
      else if ((n > 127) && (n < 2048))
         u = u sprintf("%c",_or(_rshift(n,6),192)) \
			sprintf("%c",_or(_and(n,63),128))
      else u = u sprintf("%c",_or(_rshift(n,12),224)) \
		sprintf("%c",_or(_and(_rshift(n,6),63),128)) \
				sprintf("%c",_or(_and(n,63),128))
  }

  return u
}

# =====================================================================
# string _u8decode(string u)
#
# Decode a UTF-8 encoded string.
#
# Original code from:
# http://www.webtoolkit.info/javascript-url-decode-encode.html
#
# Ported to SWU by Carlo Strozzi, September 16th, 2008
# =====================================================================

function _u8decode(u,	a,s,i,n,n2,n3) {
  n2 = n3 = 0
  i = split(u,a,"") 
  for (j=1; j<=i; j++) {
      n = _atoi(a[j])

      if (n < 128) s = s a[j]
      else if ((n > 191) && (n < 224)) {
	 n2 = _atoi(a[j+1])
	 s = s sprintf("%c", _or(_lshift(_and(n,31),6),_and(n2,63)))
	 j++
      }
      else {
	 n2 = _atoi(a[j+1])
	 n3 = _atoi(a[j+2])
	 s = s sprintf("%c",_or(_or(_lshift(_and(n,15),12),_lshift(_and(n2,63),6)),_and(n3,63)))
	 j += 2
      }
  }

  return s
}

# =====================================================================
# boolean _isutf8(string s)
#
# Return _TRUE if the received string is UTF-8 encoded, or _FALSE
# otherwise. This function is mainly for backward-compatibility.
# =====================================================================

function _isutf8(s) {

  if (_testchar(s) == "utf-8") return _TRUE

  return _FALSE
}

# =====================================================================
# void _utf8map(boolean mode)
#
# Load the UTF-8 translation lookup table. The "mode" argument can
# be either "_TRUE" "_FALSE", depending on whether we want to set the
# UTF-8 extensions or rather clear them. This function is mainly for
# backward-compatibility.
# =====================================================================

function _utf8map(mode) {

  if (mode == _TRUE) _setchar("utf-8")
  else _setchar("iso-8859-1")
}

# =====================================================================
# boolean _testchar(string s)
#
# Return the character set of the received string. This can be either
# "utf-8" or "iso-8859-1" (default). The difference between the various
# charsets in the iso-8859-* family is just a matter of how the relevant
# glyphs are displayed in the visualization environment, so it is not
# relevant to the SWU conversion lookup tables.
#
# Notes on UTF-8 detection:
#
# 1st byte of an UTF-8 char are 0xC0 to 0xFD (dec. 192 to 253)
# 2nd byte and up of an UTF-8 char are 0x80 to 0xBF (dec. 128 to 191)
#
# See http://www.cl.cam.ac.uk/~mgk25/unicode.html for more on UCS and
# UTF-8 on Linux.
# =====================================================================

function _testchar(s,					   i,j,k,a,e) {

  i = split(s,a,_NULL)

  _CHARSIZE = 1					# will be re-computed.

  # Examine one character at a time.
  for (j=1; j<=i; j++) {
      if ((e=_atoi(a[j])) > 191 && e < 254) {

	 # Looks like UTF-8, let's check next byte(s).

	 if (_CHARSIZE == 1) {
	    if (e > 251) _CHARSIZE = 6
	    else if (e > 247) _CHARSIZE = 5
	    else if (e > 239) _CHARSIZE = 4
	    else if (e > 223) _CHARSIZE = 3
	    else _CHARSIZE = 2			# minimum UTF-8 size.
	 }

	 for (k=j+1; k < (j+_CHARSIZE); k++)
	     if ((e=_atoi(a[k])) < 128 || e > 191) break

	 # Return UTF-8 upon the first complete UTF-8 character.
	 # The character size will remain set accordingly.

	 if (k == (j+_CHARSIZE)) return "utf-8"

	 # No valid UTF-8 char found, so even the computed
	 # char size must be junked and re-computed upon
	 # the next seemingly UTF-8 character.

	 _CHARSIZE = 1				# restore default.
      }
  }

  return "iso-8859-1"				# this is questionable.
}

# =====================================================================
# void _setchar(string charset)
#
# Set the specified character set. Currently supported sets are:
#
# * US-ASCII (default)
# * ISO-8859-1
# * UTF-8
#
# If a null/unknown character set is specified then US-ASCII is used.
# =====================================================================

function _setchar(name) {

  name = tolower(name)

  if (name == "iso-8859-1") {

     # Always restore US-ASCI first.
     _setchar()

     # ISO-8859-1 forward mapping.

     _CHARMAP["A0"] = _CHARMAP["a0"] = "\240"
     _CHARMAP["A1"] = _CHARMAP["a1"] = "\241"
     _CHARMAP["A2"] = _CHARMAP["a2"] = "\242"
     _CHARMAP["A3"] = _CHARMAP["a3"] = "\243"
     _CHARMAP["A4"] = _CHARMAP["a4"] = "\244"
     _CHARMAP["A5"] = _CHARMAP["a5"] = "\245"
     _CHARMAP["A6"] = _CHARMAP["a6"] = "\246"
     _CHARMAP["A7"] = _CHARMAP["a7"] = "\247"
     _CHARMAP["A8"] = _CHARMAP["a8"] = "\250"
     _CHARMAP["A9"] = _CHARMAP["a9"] = "\251"
     _CHARMAP["AA"] = _CHARMAP["aa"] = "\252"
     _CHARMAP["AB"] = _CHARMAP["ab"] = "\253"
     _CHARMAP["AC"] = _CHARMAP["ac"] = "\254"
     _CHARMAP["AD"] = _CHARMAP["ad"] = "\255"
     _CHARMAP["AE"] = _CHARMAP["ae"] = "\256"
     _CHARMAP["AF"] = _CHARMAP["af"] = "\257"
     _CHARMAP["B0"] = _CHARMAP["b0"] = "\260"
     _CHARMAP["B1"] = _CHARMAP["b1"] = "\261"
     _CHARMAP["B2"] = _CHARMAP["b2"] = "\262"
     _CHARMAP["B3"] = _CHARMAP["b3"] = "\263"
     _CHARMAP["B4"] = _CHARMAP["b4"] = "\264"
     _CHARMAP["B5"] = _CHARMAP["b5"] = "\265"
     _CHARMAP["B6"] = _CHARMAP["b6"] = "\266"
     _CHARMAP["B7"] = _CHARMAP["b7"] = "\267"
     _CHARMAP["B8"] = _CHARMAP["b8"] = "\270"
     _CHARMAP["B9"] = _CHARMAP["b9"] = "\271"
     _CHARMAP["BA"] = _CHARMAP["ba"] = "\272"
     _CHARMAP["BB"] = _CHARMAP["bb"] = "\273"
     _CHARMAP["BC"] = _CHARMAP["bc"] = "\274"
     _CHARMAP["BD"] = _CHARMAP["bd"] = "\275"
     _CHARMAP["BE"] = _CHARMAP["be"] = "\276"
     _CHARMAP["BF"] = _CHARMAP["bf"] = "\277"
     _CHARMAP["C0"] = _CHARMAP["c0"] = "\300"
     _CHARMAP["C1"] = _CHARMAP["c1"] = "\301"
     _CHARMAP["C2"] = _CHARMAP["c2"] = "\302"
     _CHARMAP["C3"] = _CHARMAP["c3"] = "\303"
     _CHARMAP["C4"] = _CHARMAP["c4"] = "\304"
     _CHARMAP["C5"] = _CHARMAP["c5"] = "\305"
     _CHARMAP["C6"] = _CHARMAP["c6"] = "\306"
     _CHARMAP["C7"] = _CHARMAP["c7"] = "\307"
     _CHARMAP["C8"] = _CHARMAP["c8"] = "\310"
     _CHARMAP["C9"] = _CHARMAP["c9"] = "\311"
     _CHARMAP["CA"] = _CHARMAP["ca"] = "\312"
     _CHARMAP["CB"] = _CHARMAP["cb"] = "\313"
     _CHARMAP["CC"] = _CHARMAP["cc"] = "\314"
     _CHARMAP["CD"] = _CHARMAP["cd"] = "\315"
     _CHARMAP["CE"] = _CHARMAP["ce"] = "\316"
     _CHARMAP["CF"] = _CHARMAP["cf"] = "\317"
     _CHARMAP["D0"] = _CHARMAP["d0"] = "\320"
     _CHARMAP["D1"] = _CHARMAP["d1"] = "\321"
     _CHARMAP["D2"] = _CHARMAP["d2"] = "\322"
     _CHARMAP["D3"] = _CHARMAP["d3"] = "\323"
     _CHARMAP["D4"] = _CHARMAP["d4"] = "\324"
     _CHARMAP["D5"] = _CHARMAP["d5"] = "\325"
     _CHARMAP["D6"] = _CHARMAP["d6"] = "\326"
     _CHARMAP["D7"] = _CHARMAP["d7"] = "\327"
     _CHARMAP["D8"] = _CHARMAP["d8"] = "\330"
     _CHARMAP["D9"] = _CHARMAP["d9"] = "\331"
     _CHARMAP["DA"] = _CHARMAP["da"] = "\332"
     _CHARMAP["DB"] = _CHARMAP["db"] = "\333"
     _CHARMAP["DC"] = _CHARMAP["dc"] = "\334"
     _CHARMAP["DD"] = _CHARMAP["dd"] = "\335"
     _CHARMAP["DE"] = _CHARMAP["de"] = "\336"
     _CHARMAP["DF"] = _CHARMAP["df"] = "\337"
     _CHARMAP["E0"] = _CHARMAP["e0"] = "\340"
     _CHARMAP["E1"] = _CHARMAP["e1"] = "\341"
     _CHARMAP["E2"] = _CHARMAP["e2"] = "\342"
     _CHARMAP["E3"] = _CHARMAP["e3"] = "\343"
     _CHARMAP["E4"] = _CHARMAP["e4"] = "\344"
     _CHARMAP["E5"] = _CHARMAP["e5"] = "\345"
     _CHARMAP["E6"] = _CHARMAP["e6"] = "\346"
     _CHARMAP["E7"] = _CHARMAP["e7"] = "\347"
     _CHARMAP["E8"] = _CHARMAP["e8"] = "\350"
     _CHARMAP["E9"] = _CHARMAP["e9"] = "\351"
     _CHARMAP["EA"] = _CHARMAP["ea"] = "\352"
     _CHARMAP["EB"] = _CHARMAP["eb"] = "\353"
     _CHARMAP["EC"] = _CHARMAP["ec"] = "\354"
     _CHARMAP["ED"] = _CHARMAP["ed"] = "\355"
     _CHARMAP["EE"] = _CHARMAP["ee"] = "\356"
     _CHARMAP["EF"] = _CHARMAP["ef"] = "\357"
     _CHARMAP["F0"] = _CHARMAP["f0"] = "\360"
     _CHARMAP["F1"] = _CHARMAP["f1"] = "\361"
     _CHARMAP["F2"] = _CHARMAP["f2"] = "\362"
     _CHARMAP["F3"] = _CHARMAP["f3"] = "\363"
     _CHARMAP["F4"] = _CHARMAP["f4"] = "\364"
     _CHARMAP["F5"] = _CHARMAP["f5"] = "\365"
     _CHARMAP["F6"] = _CHARMAP["f6"] = "\366"
     _CHARMAP["F7"] = _CHARMAP["f7"] = "\367"
     _CHARMAP["F8"] = _CHARMAP["f8"] = "\370"
     _CHARMAP["F9"] = _CHARMAP["f9"] = "\371"
     _CHARMAP["FA"] = _CHARMAP["fa"] = "\372"
     _CHARMAP["FB"] = _CHARMAP["fb"] = "\373"
     _CHARMAP["FC"] = _CHARMAP["fc"] = "\374"
     _CHARMAP["FD"] = _CHARMAP["fd"] = "\375"
     _CHARMAP["FE"] = _CHARMAP["fe"] = "\376"	# never used in UTF-8
     _CHARMAP["FF"] = _CHARMAP["ff"] = "\377"	# never used in UTF-8

     # ISO-8859-1 reverse mapping.

     _MAPCHAR["\240"] = "A0"
     _MAPCHAR["\241"] = "A1"
     _MAPCHAR["\242"] = "A2"
     _MAPCHAR["\243"] = "A3"
     _MAPCHAR["\244"] = "A4"
     _MAPCHAR["\245"] = "A5"
     _MAPCHAR["\246"] = "A6"
     _MAPCHAR["\247"] = "A7"
     _MAPCHAR["\250"] = "A8"
     _MAPCHAR["\251"] = "A9"
     _MAPCHAR["\252"] = "AA"
     _MAPCHAR["\253"] = "AB"
     _MAPCHAR["\254"] = "AC"
     _MAPCHAR["\255"] = "AD"
     _MAPCHAR["\256"] = "AE"
     _MAPCHAR["\257"] = "AF"
     _MAPCHAR["\260"] = "B0"
     _MAPCHAR["\261"] = "B1"
     _MAPCHAR["\262"] = "B2"
     _MAPCHAR["\263"] = "B3"
     _MAPCHAR["\264"] = "B4"
     _MAPCHAR["\265"] = "B5"
     _MAPCHAR["\266"] = "B6"
     _MAPCHAR["\267"] = "B7"
     _MAPCHAR["\270"] = "B8"
     _MAPCHAR["\271"] = "B9"
     _MAPCHAR["\272"] = "BA"
     _MAPCHAR["\273"] = "BB"
     _MAPCHAR["\274"] = "BC"
     _MAPCHAR["\275"] = "BD"
     _MAPCHAR["\276"] = "BE"
     _MAPCHAR["\277"] = "BF"
     _MAPCHAR["\300"] = "C0"
     _MAPCHAR["\301"] = "C1"
     _MAPCHAR["\302"] = "C2"
     _MAPCHAR["\303"] = "C3"
     _MAPCHAR["\304"] = "C4"
     _MAPCHAR["\305"] = "C5"
     _MAPCHAR["\306"] = "C6"
     _MAPCHAR["\307"] = "C7"
     _MAPCHAR["\310"] = "C8"
     _MAPCHAR["\311"] = "C9"
     _MAPCHAR["\312"] = "CA"
     _MAPCHAR["\313"] = "CB"
     _MAPCHAR["\314"] = "CC"
     _MAPCHAR["\315"] = "CD"
     _MAPCHAR["\316"] = "CE"
     _MAPCHAR["\317"] = "CF"
     _MAPCHAR["\320"] = "D0"
     _MAPCHAR["\321"] = "D1"
     _MAPCHAR["\322"] = "D2"
     _MAPCHAR["\323"] = "D3"
     _MAPCHAR["\324"] = "D4"
     _MAPCHAR["\325"] = "D5"
     _MAPCHAR["\326"] = "D6"
     _MAPCHAR["\327"] = "D7"
     _MAPCHAR["\330"] = "D8"
     _MAPCHAR["\331"] = "D9"
     _MAPCHAR["\332"] = "DA"
     _MAPCHAR["\333"] = "DB"
     _MAPCHAR["\334"] = "DC"
     _MAPCHAR["\335"] = "DD"
     _MAPCHAR["\336"] = "DE"
     _MAPCHAR["\337"] = "DF"
     _MAPCHAR["\340"] = "E0"
     _MAPCHAR["\341"] = "E1"
     _MAPCHAR["\342"] = "E2"
     _MAPCHAR["\343"] = "E3"
     _MAPCHAR["\344"] = "E4"
     _MAPCHAR["\345"] = "E5"
     _MAPCHAR["\346"] = "E6"
     _MAPCHAR["\347"] = "E7"
     _MAPCHAR["\350"] = "E8"
     _MAPCHAR["\351"] = "E9"
     _MAPCHAR["\352"] = "EA"
     _MAPCHAR["\353"] = "EB"
     _MAPCHAR["\354"] = "EC"
     _MAPCHAR["\355"] = "ED"
     _MAPCHAR["\356"] = "EE"
     _MAPCHAR["\357"] = "EF"
     _MAPCHAR["\360"] = "F0"
     _MAPCHAR["\361"] = "F1"
     _MAPCHAR["\362"] = "F2"
     _MAPCHAR["\363"] = "F3"
     _MAPCHAR["\364"] = "F4"
     _MAPCHAR["\365"] = "F5"
     _MAPCHAR["\366"] = "F6"
     _MAPCHAR["\367"] = "F7"
     _MAPCHAR["\370"] = "F8"
     _MAPCHAR["\371"] = "F9"
     _MAPCHAR["\372"] = "FA"
     _MAPCHAR["\373"] = "FB"
     _MAPCHAR["\374"] = "FC"
     _MAPCHAR["\375"] = "FD"
     _MAPCHAR["\376"] = "FE"
     _MAPCHAR["\377"] = "FF"
  }

  # No other ISO-8859-* charset should ever be necessary (see the comments
  # above). Should it ever be needed, it can be loaded as an extension to
  # the basic ISO-8859-1 charset.
  #
  #else if (name == "iso-8859-15") {
  #
  #    _setchar("iso-8859-1")
  #
  #    _CHARMAP ...
  #    _CHARMAP ...
  #
  #}

  else if (name == "utf-8") {

     # UTF-8 is always loaded only as an extension
     # of whatever previous ISO-8859-* charset.

     _setchar("iso-8859-1")

     # Set UTF-8 forward mapping.

     _CHARMAP["80"] = "\200"
     _CHARMAP["81"] = "\201"
     _CHARMAP["82"] = "\202"
     _CHARMAP["83"] = "\203"
     _CHARMAP["84"] = "\204"
     _CHARMAP["85"] = "\205"
     _CHARMAP["86"] = "\206"
     _CHARMAP["87"] = "\207"
     _CHARMAP["88"] = "\210"
     _CHARMAP["89"] = "\211"
     _CHARMAP["8A"] = _CHARMAP["8a"] = "\212"
     _CHARMAP["8B"] = _CHARMAP["8b"] = "\213"
     _CHARMAP["8C"] = _CHARMAP["8c"] = "\214"
     _CHARMAP["8D"] = _CHARMAP["8d"] = "\215"
     _CHARMAP["8E"] = _CHARMAP["8e"] = "\216"
     _CHARMAP["8F"] = _CHARMAP["8f"] = "\217"
     _CHARMAP["90"] = "\220"
     _CHARMAP["91"] = "\221"
     _CHARMAP["92"] = "\222"
     _CHARMAP["93"] = "\223"
     _CHARMAP["94"] = "\224"
     _CHARMAP["95"] = "\225"
     _CHARMAP["96"] = "\226"
     _CHARMAP["97"] = "\227"
     _CHARMAP["98"] = "\230"
     _CHARMAP["99"] = "\231"
     _CHARMAP["9A"] = _CHARMAP["9a"] = "\232"
     _CHARMAP["9B"] = _CHARMAP["9b"] = "\233"
     _CHARMAP["9C"] = _CHARMAP["9c"] = "\234"
     _CHARMAP["9D"] = _CHARMAP["9d"] = "\235"
     _CHARMAP["9E"] = _CHARMAP["9e"] = "\236"
     _CHARMAP["9F"] = _CHARMAP["9f"] = "\237"

     # Set UTF-8 reverse mapping.

     _MAPCHAR["\200"] = "80"
     _MAPCHAR["\201"] = "81"
     _MAPCHAR["\202"] = "82"
     _MAPCHAR["\203"] = "83"
     _MAPCHAR["\204"] = "84"
     _MAPCHAR["\205"] = "85"
     _MAPCHAR["\206"] = "86"
     _MAPCHAR["\207"] = "87"
     _MAPCHAR["\210"] = "88"
     _MAPCHAR["\211"] = "89"
     _MAPCHAR["\212"] = "8A"
     _MAPCHAR["\213"] = "8B"
     _MAPCHAR["\214"] = "8C"
     _MAPCHAR["\215"] = "8D"
     _MAPCHAR["\216"] = "8E"
     _MAPCHAR["\217"] = "8F"
     _MAPCHAR["\220"] = "90"
     _MAPCHAR["\221"] = "91"
     _MAPCHAR["\222"] = "92"
     _MAPCHAR["\223"] = "93"
     _MAPCHAR["\224"] = "94"
     _MAPCHAR["\225"] = "95"
     _MAPCHAR["\226"] = "96"
     _MAPCHAR["\227"] = "97"
     _MAPCHAR["\230"] = "98"
     _MAPCHAR["\231"] = "99"
     _MAPCHAR["\232"] = "9A"
     _MAPCHAR["\233"] = "9B"
     _MAPCHAR["\234"] = "9C"
     _MAPCHAR["\235"] = "9D"
     _MAPCHAR["\236"] = "9E"
     _MAPCHAR["\237"] = "9F"
  }

  else {

     # US-ASCII forward mapping.

     delete _CHARMAP

     _CHARMAP["01"] = "\001"
     _CHARMAP["02"] = "\002"
     _CHARMAP["03"] = "\003"
     _CHARMAP["04"] = "\004"
     _CHARMAP["05"] = "\005"
     _CHARMAP["06"] = "\006"
     _CHARMAP["07"] = "\007"
     _CHARMAP["08"] = "\010"
     _CHARMAP["09"] = "\011"
     _CHARMAP["0A"] = _CHARMAP["0a"] = "\012"
     _CHARMAP["0B"] = _CHARMAP["0b"] = "\013"
     _CHARMAP["0C"] = _CHARMAP["0c"] = "\014"
     _CHARMAP["0D"] = _CHARMAP["0d"] = "\015"
     _CHARMAP["0E"] = _CHARMAP["0e"] = "\016"
     _CHARMAP["0F"] = _CHARMAP["0f"] = "\017"
     _CHARMAP["10"] = "\020"
     _CHARMAP["11"] = "\021"
     _CHARMAP["12"] = "\022"
     _CHARMAP["13"] = "\023"
     _CHARMAP["14"] = "\024"
     _CHARMAP["15"] = "\025"
     _CHARMAP["16"] = "\026"
     _CHARMAP["17"] = "\027"
     _CHARMAP["18"] = "\030"
     _CHARMAP["19"] = "\031"
     _CHARMAP["1A"] = _CHARMAP["1a"] = "\032"
     _CHARMAP["1B"] = _CHARMAP["1b"] = "\033"
     _CHARMAP["1C"] = _CHARMAP["1c"] = "\034"
     _CHARMAP["1D"] = _CHARMAP["1d"] = "\035"
     _CHARMAP["1E"] = _CHARMAP["1e"] = "\036"
     _CHARMAP["1F"] = _CHARMAP["1f"] = "\037"
     _CHARMAP["20"] = "\040"
     _CHARMAP["21"] = "\041"
     _CHARMAP["22"] = "\042"
     _CHARMAP["23"] = "\043"
     _CHARMAP["24"] = "\044"
     _CHARMAP["25"] = "\045"
     _CHARMAP["26"] = "\046"
     _CHARMAP["27"] = "\047"
     _CHARMAP["28"] = "\050"
     _CHARMAP["29"] = "\051"
     _CHARMAP["2A"] = _CHARMAP["2a"] = "\052"
     _CHARMAP["2B"] = _CHARMAP["2b"] = "\053"
     _CHARMAP["2C"] = _CHARMAP["2c"] = "\054"
     _CHARMAP["2D"] = _CHARMAP["2d"] = "\055"
     _CHARMAP["2E"] = _CHARMAP["2e"] = "\056"
     _CHARMAP["2F"] = _CHARMAP["2f"] = "\057"
     _CHARMAP["30"] = "\060"
     _CHARMAP["31"] = "\061"
     _CHARMAP["32"] = "\062"
     _CHARMAP["33"] = "\063"
     _CHARMAP["34"] = "\064"
     _CHARMAP["35"] = "\065"
     _CHARMAP["36"] = "\066"
     _CHARMAP["37"] = "\067"
     _CHARMAP["38"] = "\070"
     _CHARMAP["39"] = "\071"
     _CHARMAP["3A"] = _CHARMAP["3a"] = "\072"
     _CHARMAP["3B"] = _CHARMAP["3b"] = "\073"
     _CHARMAP["3C"] = _CHARMAP["3c"] = "\074"
     _CHARMAP["3D"] = _CHARMAP["3d"] = "\075"
     _CHARMAP["3E"] = _CHARMAP["3e"] = "\076"
     _CHARMAP["3F"] = _CHARMAP["3f"] = "\077"
     _CHARMAP["40"] = "\100"
     _CHARMAP["41"] = "\101"
     _CHARMAP["42"] = "\102"
     _CHARMAP["43"] = "\103"
     _CHARMAP["44"] = "\104"
     _CHARMAP["45"] = "\105"
     _CHARMAP["46"] = "\106"
     _CHARMAP["47"] = "\107"
     _CHARMAP["48"] = "\110"
     _CHARMAP["49"] = "\111"
     _CHARMAP["4A"] = _CHARMAP["4a"] = "\112"
     _CHARMAP["4B"] = _CHARMAP["4b"] = "\113"
     _CHARMAP["4C"] = _CHARMAP["4c"] = "\114"
     _CHARMAP["4D"] = _CHARMAP["4d"] = "\115"
     _CHARMAP["4E"] = _CHARMAP["4e"] = "\116"
     _CHARMAP["4F"] = _CHARMAP["4f"] = "\117"
     _CHARMAP["50"] = "\120"
     _CHARMAP["51"] = "\121"
     _CHARMAP["52"] = "\122"
     _CHARMAP["53"] = "\123"
     _CHARMAP["54"] = "\124"
     _CHARMAP["55"] = "\125"
     _CHARMAP["56"] = "\126"
     _CHARMAP["57"] = "\127"
     _CHARMAP["58"] = "\130"
     _CHARMAP["59"] = "\131"
     _CHARMAP["5A"] = _CHARMAP["5a"] = "\132"
     _CHARMAP["5B"] = _CHARMAP["5b"] = "\133"
     _CHARMAP["5C"] = _CHARMAP["5c"] = "\134"
     _CHARMAP["5D"] = _CHARMAP["5d"] = "\135"
     _CHARMAP["5E"] = _CHARMAP["5e"] = "\136"
     _CHARMAP["5F"] = _CHARMAP["5f"] = "\137"
     _CHARMAP["60"] = "\140"
     _CHARMAP["61"] = "\141"
     _CHARMAP["62"] = "\142"
     _CHARMAP["63"] = "\143"
     _CHARMAP["64"] = "\144"
     _CHARMAP["65"] = "\145"
     _CHARMAP["66"] = "\146"
     _CHARMAP["67"] = "\147"
     _CHARMAP["68"] = "\150"
     _CHARMAP["69"] = "\151"
     _CHARMAP["6A"] = _CHARMAP["6a"] = "\152"
     _CHARMAP["6B"] = _CHARMAP["6b"] = "\153"
     _CHARMAP["6C"] = _CHARMAP["6c"] = "\154"
     _CHARMAP["6D"] = _CHARMAP["6d"] = "\155"
     _CHARMAP["6E"] = _CHARMAP["6e"] = "\156"
     _CHARMAP["6F"] = _CHARMAP["6f"] = "\157"
     _CHARMAP["70"] = "\160"
     _CHARMAP["71"] = "\161"
     _CHARMAP["72"] = "\162"
     _CHARMAP["73"] = "\163"
     _CHARMAP["74"] = "\164"
     _CHARMAP["75"] = "\165"
     _CHARMAP["76"] = "\166"
     _CHARMAP["77"] = "\167"
     _CHARMAP["78"] = "\170"
     _CHARMAP["79"] = "\171"
     _CHARMAP["7A"] = _CHARMAP["7a"] = "\172"
     _CHARMAP["7B"] = _CHARMAP["7b"] = "\173"
     _CHARMAP["7C"] = _CHARMAP["7c"] = "\174"
     _CHARMAP["7D"] = _CHARMAP["7d"] = "\175"
     _CHARMAP["7E"] = _CHARMAP["7e"] = "\176"
     _CHARMAP["7F"] = _CHARMAP["7f"] = "\177"

     # US-ASCII reverse mapping.

     delete _MAPCHAR

     _MAPCHAR["\001"] = "01"
     _MAPCHAR["\002"] = "02"
     _MAPCHAR["\003"] = "03"
     _MAPCHAR["\004"] = "04"
     _MAPCHAR["\005"] = "05"
     _MAPCHAR["\006"] = "06"
     _MAPCHAR["\007"] = "07"
     _MAPCHAR["\010"] = "08"
     _MAPCHAR["\011"] = "09"
     _MAPCHAR["\012"] = "0A"		# consider disabling.
     _MAPCHAR["\013"] = "0B"
     _MAPCHAR["\014"] = "0C"
     _MAPCHAR["\015"] = "0D"		# consider disabling.
     _MAPCHAR["\016"] = "0E"
     _MAPCHAR["\017"] = "0F"
     _MAPCHAR["\020"] = "10"
     _MAPCHAR["\021"] = "11"
     _MAPCHAR["\022"] = "12"
     _MAPCHAR["\023"] = "13"
     _MAPCHAR["\024"] = "14"
     _MAPCHAR["\025"] = "15"
     _MAPCHAR["\026"] = "16"
     _MAPCHAR["\027"] = "17"
     _MAPCHAR["\030"] = "18"
     _MAPCHAR["\031"] = "19"
     _MAPCHAR["\032"] = "1A"
     _MAPCHAR["\033"] = "1B"
     _MAPCHAR["\034"] = "1C"
     _MAPCHAR["\035"] = "1D"
     _MAPCHAR["\036"] = "1E"
     _MAPCHAR["\037"] = "1F"
     _MAPCHAR["\040"] = "20"
     _MAPCHAR["\041"] = "21"
     _MAPCHAR["\042"] = "22"
     _MAPCHAR["\043"] = "23"
     _MAPCHAR["\044"] = "24"
     _MAPCHAR["\045"] = "25"		# consider disabling.
     _MAPCHAR["\046"] = "26"
     _MAPCHAR["\047"] = "27"
     _MAPCHAR["\050"] = "28"
     _MAPCHAR["\051"] = "29"
     _MAPCHAR["\052"] = "2A"
     _MAPCHAR["\053"] = "2B"		# consider disabling.
     _MAPCHAR["\054"] = "2C"
     _MAPCHAR["\055"] = "2D"
     _MAPCHAR["\056"] = "2E"
     _MAPCHAR["\057"] = "2F"
     _MAPCHAR["\060"] = "30"
     _MAPCHAR["\061"] = "31"
     _MAPCHAR["\062"] = "32"
     _MAPCHAR["\063"] = "33"
     _MAPCHAR["\064"] = "34"
     _MAPCHAR["\065"] = "35"
     _MAPCHAR["\066"] = "36"
     _MAPCHAR["\067"] = "37"
     _MAPCHAR["\070"] = "38"
     _MAPCHAR["\071"] = "39"
     _MAPCHAR["\072"] = "3A"
     _MAPCHAR["\073"] = "3B"
     _MAPCHAR["\074"] = "3C"
     _MAPCHAR["\075"] = "3D"
     _MAPCHAR["\076"] = "3E"
     _MAPCHAR["\077"] = "3F"
     _MAPCHAR["\100"] = "40"
     _MAPCHAR["\101"] = "41"
     _MAPCHAR["\102"] = "42"
     _MAPCHAR["\103"] = "43"
     _MAPCHAR["\104"] = "44"
     _MAPCHAR["\105"] = "45"
     _MAPCHAR["\106"] = "46"
     _MAPCHAR["\107"] = "47"
     _MAPCHAR["\110"] = "48"
     _MAPCHAR["\111"] = "49"
     _MAPCHAR["\112"] = "4A"
     _MAPCHAR["\113"] = "4B"
     _MAPCHAR["\114"] = "4C"
     _MAPCHAR["\115"] = "4D"
     _MAPCHAR["\116"] = "4E"
     _MAPCHAR["\117"] = "4F"
     _MAPCHAR["\120"] = "50"
     _MAPCHAR["\121"] = "51"
     _MAPCHAR["\122"] = "52"
     _MAPCHAR["\123"] = "53"
     _MAPCHAR["\124"] = "54"
     _MAPCHAR["\125"] = "55"
     _MAPCHAR["\126"] = "56"
     _MAPCHAR["\127"] = "57"
     _MAPCHAR["\130"] = "58"
     _MAPCHAR["\131"] = "59"
     _MAPCHAR["\132"] = "5A"
     _MAPCHAR["\133"] = "5B"
     _MAPCHAR["\134"] = "5C"
     _MAPCHAR["\135"] = "5D"
     _MAPCHAR["\136"] = "5E"
     _MAPCHAR["\137"] = "5F"
     _MAPCHAR["\140"] = "60"
     _MAPCHAR["\141"] = "61"
     _MAPCHAR["\142"] = "62"
     _MAPCHAR["\143"] = "63"
     _MAPCHAR["\144"] = "64"
     _MAPCHAR["\145"] = "65"
     _MAPCHAR["\146"] = "66"
     _MAPCHAR["\147"] = "67"
     _MAPCHAR["\150"] = "68"
     _MAPCHAR["\151"] = "69"
     _MAPCHAR["\152"] = "6A"
     _MAPCHAR["\153"] = "6B"
     _MAPCHAR["\154"] = "6C"
     _MAPCHAR["\155"] = "6D"
     _MAPCHAR["\156"] = "6E"
     _MAPCHAR["\157"] = "6F"
     _MAPCHAR["\160"] = "70"
     _MAPCHAR["\161"] = "71"
     _MAPCHAR["\162"] = "72"
     _MAPCHAR["\163"] = "73"
     _MAPCHAR["\164"] = "74"
     _MAPCHAR["\165"] = "75"
     _MAPCHAR["\166"] = "76"
     _MAPCHAR["\167"] = "77"
     _MAPCHAR["\170"] = "78"
     _MAPCHAR["\171"] = "79"
     _MAPCHAR["\172"] = "7A"
     _MAPCHAR["\173"] = "7B"
     _MAPCHAR["\174"] = "7C"
     _MAPCHAR["\175"] = "7D"
     _MAPCHAR["\176"] = "7E"
     _MAPCHAR["\177"] = "7F"
  }
}

# =====================================================================
# End of program
# =====================================================================
