#!/usr/local/bin/mawk -We
# *********************************************************************
# updmerge: inserts/updates/deletes table rows based on the contents
#	    of an edit table. This is a fast, one pass process that assumes
#	    the following:
#             1) Both files are sorted by primary key
#             2) All primary key values are the same length
#             3) Single segment primary keys
#             4) Both tables have identical structures
#
# Copyright (C) 2007 by Steve Litt
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 2 dated June, 1991.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# *********************************************************************
# $Id: updmerge.awk 39 2007-05-02 14:46:09Z carlo $

BEGIN{
	### HANDLE ARGUMENTS, DEFEAT DEFAULT USE OF ARGS AS INPUT FILES ###
	editfilename = ARGV[2]
	ARGC = 1

	### FILE SEPARATOR SHOULD BE TAB ###
	FS="\t"

	### SET SOME OTHER VARS ###
	record_changed_or_deleted_flag = 0
	high_value_string="\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF"

	### BLOW OFF EDIT FILE'S HEADER, READ ITS FIRST DATA RECORD ###
	### TO ENABLE THE MERGE ALGORTHM ###
	numchars = getline editline < editfilename  # Blow off edit file's header
	if(numchars == 0){
		print "Empty edit file (" editfilename ", aborting." > stderr
		exit 1
	}

	editline = ""
	numchars = getline editline < editfilename  # Get first edit line
	if(numchars == 0){
		print "No data in edit file (" editfilename ", aborting." > stderr
		exit 1
	}

	### SPLIT editline INTO FIELDS FOR COMPARISON WITH MAIN FILE RECORD ###
	split(editline, editarray, "\t")
}

### MAIN PROGRAM  ### NOT USED FOR ANYTHING IMPORTANT RIGHT NOW
{
	record_changed_or_deleted_flag = 0
}

### RULES
NR == 1{                  # Transfer main header intact, always
	print $0
	next
}
$1 < editarray[1]{        # If main record smaller than edit, print main
	print $0
	next
}

### If edit record is smaller or equal than main, the edit record must
### be an add if it's bigger, or a delete or change if it's equal.
### iterate through all lower or equal edit records, not quitting
### til you have an edit record bigger than the main record.
$1 >= editarray[1]{
	while($1 >= editarray[1]){
		if($1 > editarray[1]){
			print editline
		}else {
			record_changed_or_deleted_flag = 1

			### LOOK FOR DEL AND XFER DATA TO OUTPUT
			for(i=1; i <= NF; i++){ outarray[i] = $i }
			for(i=2; i <= NF; i++){
				if(i in editarray){
					if(editarray[i] ~ /^\.\.DEL\.\./){
						delflag = 1
						break
					}
					outarray[i] = editarray[i]
				}
			}
			if(!delflag) {
				### ASSEMBLE OUTPUT LINE
				output = outarray[1]
				for(i=2; i <= NF; i++){
					output = output "\t" outarray[i]
				}
				### PRINT OUTPUT LINE
				print output
			}
		}

		### GET NEXT EDIT RECORD ###
		numchars = getline editline < editfilename 
		if(numchars == 0){
			editline = high_value_string
			editarray[1]=high_value_string
		} else {
			split(editline, editarray)
		}
	}

}

### GUARANTEE PRINT FOR FIRST RECORD OF MAIN FILE ###
### BUT ONLY AFTER ANY LOWER EDIT RECORDS HAVE BEEN PRINTED ###
### AND ONLY IF THE FIRST RECORD HASN'T BEEN CHANGED OR DELETED ###
NR == 2{
	if(!record_changed_or_deleted_flag){print $0}
}

END{    # INSERT ANY EDIT RECORDS ABOVE HIGHEST MAIN RECORD
	while(editarray[1] != high_value_string){
		### DON'T PRINT DELETES ###
		delflag=0
		for(i=2; i <= NF; i++){
			if(i in editarray){
				if(editarray[i] ~ /^\.\.DEL\.\./){
					delflag = 1
					break
				}
			}
		}

		if(!delflag) print editline

		### GET NEXT EDIT TABLE RECORD ###
		numchars = getline editline < editfilename 
		if(numchars == 0){
			editline = high_value_string
			editarray[1]=high_value_string
		} else {
			split(editline, editarray)
		}
	}
}
